'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
import tempfile
import os
import pwd
import sys

################################################################################
# environment / options
################################################################################

def convert2boolean(arg):
    '''
    convert2boolean is used for environmental variables
    that must be returned as boolean
    '''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg


def getenv(variable_key, default=None, required=False, silent=True):
    '''
    getenv will attempt to get an environment variable. If the variable
    is not found, None is returned.

    :param variable_key: the variable name
    :param required: exit with error if not found
    :param silent: Do not print debugging information for variable
    '''
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        bot.error("Cannot find environment variable %s, exiting." %variable_key)
        sys.exit(1)

    if not silent and variable is not None:
            bot.verbose("%s found as %s" %(variable_key,variable))

    return variable


def getenv_namespace(namespace="SCIF", func=None):
    '''return all environment variables in a particular namespace, such as for
       SCIF apps using a filter function (func). The function should take first
       a key in the environment, and then the namespace variable. 
       If none provided, uses "starts with" equivalent.

        def func(key, namespace):
            return key.startswith(namespace)
    '''
    if func is None:
        '''does key start with namespace?'''
        def func(key, namespace):
            return key.startswith(namespace)

    env = os.environ.items()
    return [{key:value} for key,value in env if func(key,namespace)]


################################################################################
# Defaults


# Supported Sections
sections = ('appenv',
            'applabels',
            'appinstall',
            'appfiles',
            'apphelp',
            'apprun',
            'apptest')


################################################################################
# Scientific Filesystem (environment)


#########################
# Global Settings
#########################

SCIF_APPS = getenv_namespace(namespace="SCIF_APP")
SCIF_BASE = getenv('SCIF_BASE', '/scif')
SCIF_DATA = getenv('SCIF_DATA', '%s/data' %SCIF_BASE)
SCIF_APPS = getenv('SCIF_APPS', '%s/apps' %SCIF_BASE)
SCIF_PYSHELL = getenv('SCIF_PYSHELL', 'ipython')
SCIF_SHELL = getenv('SCIF_SHELL', '/bin/bash')
SCIF_ENTRYPOINT = getenv('SCIF_ENTRYPOINT','/bin/bash')
SCIF_ENTRYFOLDER = getenv('SCIF_ENTRYFOLDER') # default None, is set to /scif
                                              # in functions OR app roots

#########################
# Permissions
#########################

SCIF_APPEND_PATHS = ['PYTHONPATH', 'PATH', 'LD_LIBRARY_PATH']
SCIF_ALLOW_APPEND = convert2boolean(getenv('SCIF_ALLOW_APPEND', True))



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from collections import OrderedDict
from scif.logger import bot
from scif.utils import read_file
from scif.defaults import sections
import os
import re
import sys

def load_filesystem(base, quiet=False):
    '''load a filesystem based on a root path, which is usually /scif

    Parameters
    ==========
    base: base to load.

    Returns
    =======
    config: a parsed recipe configuration for SCIF
    '''
    from scif.defaults import SCIF_APPS

    if os.path.exists(SCIF_APPS):
        apps = os.listdir(SCIF_APPS)
        config = {'apps': {}}
        for app in apps:
            path = '%s/%s/scif/%s.scif' %(SCIF_APPS, app, app)
            if os.path.exists(path):
                recipe = load_recipe(path)
                config['apps'][app] = recipe['apps'][app]

        if len(config['apps']) > 0:
            if quiet is False:
                bot.info('Found configurations for %s scif apps' %len(config['apps']))
                bot.info('\n'.join(list(config['apps'].keys())))
            return config


def load_recipe(path):
    '''load will return a loaded in (user) scif configuration file

    Parameters
    ==========
    path: a path to a scif recipe file

    Returns
    =======
    config: a parsed recipe configuration for SCIF
    '''

    path = os.path.abspath(path)
    if os.path.exists(path):

        # Read in spec, skip empty lines, don't strip remaining
        spec = [x.strip('\n') 
               for x in read_file(path) 
               if not re.match('^\s+$', x.strip('\n'))]

        spec = [x for x in spec if x not in ['', None]]
        config = OrderedDict()
        section = None
        name = None

        while len(spec) > 0:

            # Clean up white trailing/leading space
            line = spec.pop(0)
            stripped = line.strip()

            # Comment
            if stripped.startswith("#"):
                continue

            # A new section?
            elif stripped.startswith('%'):

                # Remove any comments
                line = line.split('#',1)[0].strip()

                # Is there a section name?
                parts = line.split(' ')
                if len(parts) > 1:
                    name = ' '.join(parts[1:])          
                section = re.sub('[%]|(\s+)','',parts[0]).lower()
                config = add_section(config=config,
                                     section=section,
                                     name=name)

            # If we have a section, and are adding it
            elif section is not None:
                spec = [line] + spec
                config = read_section(config=config,
                                      spec=spec,
                                      section=section,
                                      name=name)


        # Make sure app environments are sourced as first line of recipe
        config = finish_recipe(config)   
        
    else:
        bot.debug("Cannot find recipe file %s" %path)
    return config


def finish_recipe(config, global_section='apps'):
    '''
       finish recipe includes final steps to add to the runtime for an app.
       Currently, this just means adding a command to source an environment
       before running, if appenv is defined. The Python should handle putting
       variables in the environment, however in some cases (if the variable
       includes an environment variable:

          VARIABLE1=$VARIABLE2

       It would not be properly sourced! So we add a source as the first
       line of the runscript

       Parameters
       ==========
       config: the configuation file produced by load_recipe. Assumed to have
               a highest key of "apps" and then lookup by individual apps,
               and then sections. Eg: config['apps']['myapp']['apprun'] 

    '''
    # The apps are the keys under global section "apps"
    apps = list(config[global_section].keys())

    for app in apps:

        # If an apprun is present and the system supports source, do it.
        if "appenv" in config[global_section][app]:
            appenv = config[global_section][app]['appenv']

            # If runscript is defined, add source to first line
            if "apprun" in config[global_section][app]:
                apprun = config[global_section][app]['apprun']
                config[global_section][app]['apprun'] =  appenv + apprun
            else:    
                # An app can just be an environment
                config[global_section][app]['apprun'] = appenv


    return config


def read_section(config, spec, section, name, global_section='apps'):
    '''read in a section to a list, and stop when we hit the next section
    '''
    members = []

    while True:

        if len(spec) == 0:
            break
        next_line = spec[0]                

        if next_line.upper().strip().startswith("%"):
            break
        else:
            new_member = spec.pop(0)
            if not new_member.strip().startswith('#'):

                # Strip whitespace for labels, files, environment
                if section in ['applabels', 'appfiles', 'appenv']:
                    new_member = new_member.strip()

                members.append(new_member)

    # Add the list to the config
    if len(members) > 0:
        if section is not None and name is not None:
            config[global_section][name][section] = members
        else: # section is None, is just global
            config[global_section] = members

    return config


def add_section(config, section, name=None, global_section="apps"):
    '''add section will add a section (and optionally)
    section name to a config

    Parameters
    ==========
    config: the config (dict) parsed thus far
    section: the section type (e.g., appinstall)
    name: an optional name, added as a level (e.g., google-drive)

    Resulting data structure is:

            config['registry']['apprun']
            config[name][section]

    '''

    if section is None:
        bot.error('You must define a section (e.g. %appenv) before any action.')
        sys.exit(1)

    if section not in sections:
        bot.error("%s is not a valid section." %section)
        sys.exit(1)

    # Add the global section, if doesn't exist
    if global_section not in config:
        config[global_section] = OrderedDict()

    if name is not None:        
        if name not in config[global_section]:
            config[global_section][name] = OrderedDict()

        if section not in config[global_section][name]: 
            config[global_section][name][section] = []
            bot.debug("Adding section %s %s" %(name, section))

    return config



'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
from scif.defaults import SCIF_BASE
from scif.utils import which
import locale
import sys
import os


def _exec(self, app=None, interactive=False):
    '''exec is the underlying driver of both run and exec, taking a final
       SCIF and executing the command for the user.

       This function is called via self._exec, and the expectation is that
       self.run or self.exec has been called first to prepare the environment
       and SCIF settings.

       If a user wants to preserve an environment variable from the console
       it can be referenced with [e], for example $SCIF_DATA --> [e]SCIF_DATA
    '''

    name = ''
    if app is not None and app in self.apps():
        name = '[%s] ' %app

    # If the entry folder still not set, we don't have an app
    if self._entry_folder is None:
        self._entry_folder = SCIF_BASE

    # Change directory to the relevant place
    os.chdir(self._entry_folder)

    # export environment
    runtime_environ = self.export_env() 

    # Execv to the entrypoint
    executable = which(self._entry_point[0])
    args = ''

    if len(self._entry_point) > 1:
        args = ' '.join(self._entry_point[1:])

    cmd = "%s %s" %(executable, args)
    bot.info('%sexecuting %s' %(name, cmd))

    # Return output to console
    loc = locale.getdefaultlocale()[1]

    # A shell will run the command
    if interactive is True:
        os.system(''.join(cmd))
    else:
        for line in os.popen(cmd):
            try:
                print(line.rstrip())
            except:
                print(line.rstrip().encode(loc))


def execute(self, app, cmd=None):
    '''execute a command in the context of an app. This means the following:

    1. Check that the app is valid for the client. Don't proceed otherwise
    2. Set the client app to be active
    3. update the environment to indicate the app is active
    4. set the entry point for exec to be relative to the app

    Parameters
    ==========
    app: the name of the scif app to execute a command to

    '''
    self.activate(app, cmd)
                          # checks for existence
                          # sets _active to app's name
                          # updates environment
                          # sets entrypoint
                          # sets entryfolder

    return self._exec(app)


def shell(self, app, cmd=None):
    '''akin to execute, but specific for shell. In this case, we pass the
       calling process to the shell. We do the same steps as in run/exec:

    1. Check that the app is valid for the client. Don't proceed otherwise
    2. Set the client app to be active
    3. update the environment to indicate the app is active
    4. set the entry point for exec to be relative to the app

    Parameters
    ==========
    app: the name of the scif app to execute a command to

    '''
    from scif.defaults import SCIF_SHELL

    if cmd is None:
        cmd = SCIF_SHELL

    self.activate(app, cmd)
                            # checks for existence
                            # sets _active to app's name
                            # updates environment
                            # sets shell entrypoint

                           # interactive
    return self._exec(app, True)



def run(self, app=None, args=None):
    '''run an app. This means the following:

    1. Check that the app is valid for the client. Don't proceed otherwise
    2. Set the client app to be active
    3. update the environment to indicate the app is active
    4. set the entry point and entry folder relevant to the app

    Parameters
    ==========
    app: the name of the scif app to run
    args: a list of one or more additional arguments to pass to runscript

    '''
    self.activate(app, args=args)    # checks for existence
                                     # sets _active to app's name
                                     # updates environment
                                     # sets entrypoint
                                     # sets entryfolder

    return self._exec(app)



'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.main.apps import ( app, apps, activate, deactivate, help, 
                             inspect, reset )
from scif.main.base import ScifRecipe
from scif.main.commands import ( _exec, execute, run, shell )
from scif.main.environment import ( 
    add_env, 
    append_path,
    export_env,
    get_env, 
    get_append_path,
    get_appenv,
    get_appenv_lookup, 
    init_env, 
    load_env,
    update_env
)
from scif.main.helpers import run_command
from scif.main.preview import ( 
    preview, 
    preview_apps, 
    preview_base, 
    init_app_preview,
    preview_runscript,
    preview_labels,
    preview_environment,
    preview_files,
    preview_commands,
    preview_recipe,
    preview_test
)
from scif.main.setup import ( install_base, set_base, set_defaults )
from scif.main.install import (
    init_app,
    install,
    install_apps,
    install_runscript,
    install_environment,
    install_help,
    install_script,
    install_labels,
    install_commands,
    install_files,
    install_recipe,
    install_test
)


# We can eventually add logic here for customizing the client


# Commands
ScifRecipe.execute = execute
ScifRecipe._exec = _exec
ScifRecipe.run = run
ScifRecipe.shell = shell

# Helpers
ScifRecipe._run_command = run_command
ScifRecipe.help = help

# Environment
ScifRecipe.append_path = append_path
ScifRecipe._append_path = get_append_path
ScifRecipe._init_env = init_env
ScifRecipe.add_env = add_env
ScifRecipe.export_env = export_env
ScifRecipe.get_env = get_env
ScifRecipe.load_env = load_env
ScifRecipe.update_env = update_env


# Preview
ScifRecipe.preview = preview
ScifRecipe._preview_base = preview_base
ScifRecipe._preview_apps = preview_apps
ScifRecipe._init_app_preview = init_app_preview
ScifRecipe._preview_runscript = preview_runscript
ScifRecipe._preview_labels = preview_labels
ScifRecipe._preview_environment = preview_environment
ScifRecipe._preview_commands = preview_commands
ScifRecipe._preview_files = preview_files
ScifRecipe._preview_recipe = preview_recipe
ScifRecipe._preview_test = preview_test

# Setup
ScifRecipe._install_base = install_base
ScifRecipe.set_base = set_base
ScifRecipe.set_defaults = set_defaults

# Apps
ScifRecipe.get_appenv_lookup = get_appenv_lookup
ScifRecipe.get_appenv = get_appenv
ScifRecipe.app = app
ScifRecipe.apps = apps
ScifRecipe.activate = activate
ScifRecipe.deactivate = deactivate
ScifRecipe.inspect = inspect
ScifRecipe.reset = reset

# Installation
ScifRecipe.install = install
ScifRecipe._init_app = init_app
ScifRecipe._install_apps = install_apps
ScifRecipe._install_commands = install_commands
ScifRecipe._install_environment = install_environment
ScifRecipe._install_files = install_files
ScifRecipe._install_help = install_help
ScifRecipe._install_labels = install_labels
ScifRecipe._install_recipe = install_recipe
ScifRecipe._install_runscript = install_runscript
ScifRecipe._install_script = install_script
ScifRecipe._install_test = install_test



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
import subprocess
import os
import re
import sys


# External Environment Functions

def append_path(self, varname, value):
    '''append another directory to a path indexed by "varname" in the os.environ
       these variables are typically not represented by the user in the appenv,
       but are added dynamically by SCIF. Thus, they will appear in os.environ
       groups that are added to self.environment at runtime.

       Parameters
       ==========
       varname: the variable in the environment to append to (e.g., PATH)
       value: the path to append (no :)

    '''
    value = self._append_path(varname, value)
    os.environ[varname] = value
    os.putenv(varname, value)


def get_append_path(self, key, value, environ=None):
    '''the driver function of append_path, with intention to return the correct
       value and variable name for an environment of interest, this function
       maps to _append_path e.g.:
  
       1. if the variable is already defined, we append
       2. if the variable isn't defined, we set

       Parameters
       ==========
       key: the variable in the environment to append to (e.g., PATH)
       value: the path to append (no :)
       environment: the environment to look in.

    '''
    from scif.defaults import ( SCIF_APPEND_PATHS, SCIF_ALLOW_APPEND )

    if environ is None:
        environ = os.environ

    if key in environ and SCIF_ALLOW_APPEND and key in SCIF_APPEND_PATHS:
        return "%s:%s" %(value, os.environ[key])
    return value


def init_env(self, config, base='/scif', active=None):
    '''env will parse the complete SCIF namespace environment from a config.
    this will be defined for all apps to allow for easy interaction between
    them, regardless of which individual app is active.
    
    Parameters
    ==========
    config: the config loaded in with "load" in this file.
    active: the name of the active app, if relevant.

    Example: the following environment variables would be defined for an app
             called "google-drive" Note that for the variable, the slash is
             replaced with an underscore

             SCIF_APPDATA_google_drive=/scif/data/google-drive
             SCIF_APPRUN_google_drive=/scif/apps/google-drive/scif/runscript
             SCIF_APPHELP_google_drive=/scif/apps/google-drive/scif/runscript.help
             SCIF_APPROOT_google_drive=/scif/apps/google-drive
             SCIF_APPLIB_google_drive=/scif/apps/google-drive/lib
             SCIF_APPMETA_google_drive=/scif/apps/google-drive/scif
             SCIF_APPBIN_google_drive=/scif/apps/google-drive/bin
             SCIF_APPENV_google_drive=/scif/apps/google-drive/scif/environment.sh
             SCIF_APPLABELS_google_drive=/scif/apps/google-drive/scif/labels.json

        These paths and files are not created at this point, but just defined.

    '''
    envars = dict()
    if "apps" in config:
        for name, app in config['apps'].items(): 

            # Here we are adding variables for all apps.
            appenv = self.get_appenv_lookup(name)
            for varname, varval in appenv[name].items():
                updates = mk_env(key=varname, val=varval, app=name)
                envars.update(updates)

            # If this is the active app
            if active is not None:
                if active == name:
                    for varname, varval in appenv[name].items():
                        updates = mk_env(key=varname, val=varval)
                        envars.update(updates)

    return envars


def update_env(self, reset=False):
    '''If the SCIF is loaded, upload the object's environment.

    Parameters
    ==========
    reset: if True, empty the environment before parsing from config

    '''

    if reset is True:
        self.environment = dict()

    if self._config is not None:

        # Update environment with app information
        updates = self._init_env(self._config, self._base)
        self.environment.update(updates)

    return self.environment


def mk_env(key, val, app=None):
    '''a helper function to return a dictionary with a SCIF prefixed app name,
       the key slashes replaced with underscores, and the value set.

       Parameters
       ==========
       app: the app name to define the variable for (not used if not defined)
       key: the key (not including the SCIF_ prefix
       val: the value to set

    '''
    key = "SCIF_%s" % key.upper()
    if app is not None:
        key = "%s_%s" %(key,app)

    key = re.sub('-|[.]', '_', key)
    return { key:val }


def load_env(self, app):
    '''load an environment file for an app. We don't allow for manual entry
       of any file, but limit to predetermined environment.sh file, if exists.
       this function is inteded for runtime commands (shell or exec not install)

       Parameters
       ==========
       app: the app to export variables for

    '''
    updates = dict()

    if app in self.apps():
        config = self.get_appenv(app)

        if 'SCIF_APPENV' in config:
            envfile = config['SCIF_APPENV']
            if os.path.exists(envfile):
                with open(envfile, 'r') as filey:
                    lines = filey.readlines()
                for line in lines:
                    (key, _, val) = line.strip().partition("=")
                    if val not in ['', None]: # skips export lines

                        updates[key] = val
                        self.environment[key] = val
    return updates


def export_env(self, ps1=True):
    '''export the current environment, and add the PS1 variable to indicate
       the active shell display. This will start with values from the currently
       active environment, and then add those from scif.

       Parameters
       =========
       ps1: if True, change the shell prompt to scif>

    '''  

    runtime = os.environ.copy()

    if ps1 is True:
        runtime['PS1'] = "scif> "

    if hasattr(self,'environment'):

        # Step 1. Do an update, allowing extension for PATHs
        for key,val in self.environment.items():
            runtime[key] = self._append_path(key, val, self.environment)  

        # Step 2; export
        for key,val in runtime.items():
            os.environ[key] = val
            os.putenv(key, val)

    return runtime


def get_appenv(self, app, isolated=True, update=False):
    '''return environment for a specific app, meaning the variables active
       when it is running. 

       Parameters
       ==========
       isolated: if True, only return the active app variables (example  below)
       update: also update the global self.environment.

       If isolated is True, don't include other apps. For
       example, for an app `hello-world-echo` and isolated True, the following
       is returned:

       {
           'SCIF_APPBIN': '/scif/apps/hello-world-echo/bin',
           'SCIF_APPDATA': '/scif/data/hello-world-echo',
           'SCIF_APPENV': '/scif/apps/hello-world-echo/scif/environment.sh',
           'SCIF_APPHELP': '/scif/apps/hello-world-echo/scif/runscript.help',
           'SCIF_APPLABELS': '/scif/apps/hello-world-echo/scif/labels.json',
           'SCIF_APPLIB': '/scif/apps/hello-world-echo/lib',
           'SCIF_APPMETA': '/scif/apps/hello-world-echo/scif',
           'SCIF_APPNAME': 'hello-world-echo',
           'SCIF_APPRECIPE': '/scif/apps/hello-world-echo/scif/hello-world-echo.scif',
           'SCIF_APPROOT': '/scif/apps/hello-world-echo',
           'SCIF_APPRUN': '/scif/apps/hello-world-echo/scif/runscript'
       }
    '''
    final = dict()
    if app in self.apps():
        environ = self.get_appenv_lookup(app)
        for var, val in environ[app].items():
            updates = mk_env(key=var, val=val)
            final.update(updates)

        # The user wants to include the current SCIF environment
        if isolated is False and hasattr(self,'environment'):
            final.update(self.environment)

        # The user also wants to update the SCIF environment
        if update is True:
            self.environment = final
        
        return final

    valid = ' '.join(self.apps())
    bot.error('%s is not a valid app. Found %s' %(app, valid))


def get_appenv_lookup(self, app):
    '''create a dictionary with a highest level index the
       app name, and underneath a generic lookup (without the app name) for
       different variable types.  

       Parameters
       ==========
       app: the new of the app to get the environment for
       isolated: if True don't include other apps

       Eg, app with name "google-drive" would look like:


       {'registry': {
                      'appbin': '/scif/apps/registry/bin',
                      'appdata': '/scif/data/registry',
                      'appenv': '/scif/apps/registry/scif/environment.sh',
                      'apphelp': '/scif/apps/registry/scif/runscript.help',
                      'apptest': '/scif/apps/registry/scif/test.sh',
                      'applabels': '/scif/apps/registry/scif/labels.json',
                      'applib': '/scif/apps/registry/lib',
                      'appmeta': '/scif/apps/registry/scif',
                      'apprecipe': '/scif/apps/registry/scif/registry.scif'
                      'approot': '/scif/apps/registry',
                      'apprun': '/scif/apps/registry/scif/runscript'
                    }            
       }

       This function is intended to be shared by env above and the environment
       generating functions in the main client, to have consistent behavior. 
       The above data structure gets parse into the (global) variables for
       the particular app:

       {'SCIF_APPBIN_registry': '/scif/apps/registry/bin',
        'SCIF_APPDATA_registry': '/scif/data/registry',
        'SCIF_APPENV_registry': '/scif/apps/registry/scif/environment.sh',
        'SCIF_APPHELP_registry': '/scif/apps/registry/scif/runscript.help',
        'SCIF_APPLABELS_registry': '/scif/apps/registry/scif/labels.json',
        'SCIF_APPTEST_registry': '/scif/apps/registry/scif/test.sh',
        'SCIF_APPLIB_registry': '/scif/apps/registry/lib',
        'SCIF_APPMETA_registry': '/scif/apps/registry/scif',
        'SCIF_APPRECIPE_registry': '/scif/apps/registry/scif/registry.scif',
        'SCIF_APPROOT_registry': '/scif/apps/registry',
        'SCIF_APPRUN_registry': '/scif/apps/registry/scif/runscript'}

    '''

    if app in self.apps():

        base = self._base
        envars = {app:{}}

        # Roots for app data and app files
        appdata = "%s/data/%s" %(base, app)
        approot = "%s/apps/%s" %(base, app)
        appmeta = "%s/scif"  %(approot)

        envars[app]['appdata'] = appdata
        envars[app]['approot'] = approot
        envars[app]['appmeta'] = appmeta

        envars[app]['appbin'] = "%s/bin"  %(approot)
        envars[app]['applib'] = "%s/lib"  %(approot)
        envars[app]['apprun'] = "%s/runscript"  %(appmeta)
        envars[app]['apphelp'] = "%s/runscript.help"  %(appmeta)
        envars[app]['appenv'] = "%s/environment.sh"  %(appmeta)
        envars[app]['apptest'] = "%s/test.sh"  %(appmeta)
        envars[app]['applabels'] = "%s/labels.json"  %(appmeta)
        envars[app]['apprecipe'] = "%s/%s.scif"  %(appmeta, app)
        envars[app]['appname'] = app
        return envars

    # if we get down here, didn't have the app in the first place
    valid = ' '.join(self.apps())
    bot.error('%s is not a valid app. Found %s' %(app, valid))


# ScifRecipe Environment Helpers

def add_env(self, key, value):
    '''add a key/value pair to the environment. Should begin with SCIF
       to maintain proper namespace

    Parameters
    ==========
    key: the environment variable name. For SCIF, slashes should be 
         removed and replaced with underscore.
    value: the value to set for the environment variable

    '''    
    if not hasattr(self,'environment'):
        self.environment = dict()

    if not key.startswith('SCIF'):
        msg = 'Environment variable outside SCIF namespace not recommended.'
        bot.warning(msg)

    # If the variable already exists, status is "update"
    action = 'new'
    if key in self.environment:
        action = 'update'

    self.environment[key] = value
    bot.debug('[environment:%s][%s=%s]' %(action, key, value))


def get_env(self, key=None):
    '''return a particular value for an environment variable, if the key
       exists. If no keys are defined, the entire environment is returned.
       For an app-specific environment, use get_appenv.
    '''
    if key is not None:
       if key in self.environment:
           bot.info(self.environment[key])
           return self.environment[key]

    # otherwise, return the entire environment lookup
    if hasattr(self, 'environment'):
        return self.environment



'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''


from scif.logger import bot
from datetime import datetime
from scif.utils import ( mkdir_p, run_command, write_file, write_json )
from scif.main.helpers import get_parts
import sys
import os



def preview(self, apps=None):
    '''preview the complete setup for a scientific filesytem. This is useful 
       to print out actions for install (without doing them).
    '''
    self._preview_base()             # preview the folder structure
    self._preview_apps(apps)          # App install


def preview_base(self):
    ''' preview basic scif structure at the base for apps and metadata

        Parameters
        ==========
        base: the full path to the root folder to create /scif
    '''
    if not hasattr(self,'_base'):
        bot.error('Please set the base before preview.')
        sys.exit(1)
    
    bot.custom(prefix='[base] %s' %self._base, color='CYAN')  
    bot.custom(prefix='[apps] %s' %self.path_apps, color='CYAN')  
    bot.custom(prefix='[data] %s\n' %self.path_data, color='CYAN')


def preview_apps(self, apps=None):
    '''install one or more apps to the base. If app is defined, only
       install app specified. Otherwise, install all found in config.
    '''
    if apps in [None,[]]:
        apps = self.apps()

    if not isinstance(apps, list):
        apps = [apps]


    for app in apps:

        # We must have the app defined in the config
        if app not in self._config['apps']:
            bot.error('Cannot find app %s in config.' %app)
            sys.exit(1)

        # Make directories
        bot.newline()
        settings = self._init_app_preview(app)

        # Get the app configuration
        config = self.app(app)

        # Handle environment, runscript, labels
        self._preview_runscript(app, settings, config)
        self._preview_environment(app, settings, config)
        self._preview_labels(app, settings, config)
        self._preview_commands(app, settings, config)
        self._preview_files(app, settings, config)
        self._preview_recipe(app, settings, config)
        self._preview_test(app, settings, config)
        bot.newline()

def init_app_preview(self, app):
    '''initialize an app, meaning adding the metadata folder, bin, and 
       lib to it. The app is created at the base
    '''
    settings = self.get_appenv_lookup(app)[app]

    for key in ['root', 'lib', 'bin', 'data']:
        val = settings['app%s' %key]
        bot.custom(prefix="[%s] %s" %(key, val), color='CYAN')
    
    return settings


def preview_runscript(self, app, settings, config):
    '''preview the runscript for an app
       
       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    if "apprun" in config:
        content = '\n'.join(config['apprun'])
        bot.info('+ ' + 'apprun '.ljust(5) + app)
        print(settings['apprun'])
        print(settings['apphelp'])
        print(content)


def preview_test(self, app, settings, config):
    '''preview the test for the app

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    if "apptest" in config:
        content = '\n'.join(config['apptest'])
        bot.info('+ ' + 'apptest '.ljust(5) + app)
        print(settings['apptest'])
        print(content)


def preview_labels(self, app, settings, config):
    '''preview labels for an app

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    lookup = ''
    if "applabels" in config:
        labels = config['applabels']
        bot.info('+ ' + 'applabels '.ljust(5) + app)
        print(settings['applabels'])
        for line in labels:
            label, value = get_parts(line, default='')
            lookup += '%s=%s\n' %(label,value)
        print(lookup)
    return lookup


def preview_files(self, app, settings, config):
    '''install files will add files (or directories) to a destination.
       If none specified, they are placed in the app base

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''

    if "appfiles" in config:
        files = config['appfiles']
        bot.info('+ ' + 'appfiles '.ljust(5) + app)

        for pair in files:

            # Step 1: determine source and destination
            src, dest = get_parts(pair, default=settings['approot'])
            print('%s --> %s \n' %(src,dest)) 


def preview_commands(self, app, settings, config):
    '''install will finally, issue commands to install the app.

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    cmd = ''
    if "appinstall" in config:
        cmd = config['appinstall']
        bot.info('+ ' + 'appinstall '.ljust(5) + app)
        print('\n'.join(cmd))
    return cmd


def preview_recipe(self, app, settings, config):
    '''Write the initial recipe for the app to its metadata folder.

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    recipe_file = settings['apprecipe']
    recipe = '# [scif] scientific filesystem recipe\n' 
    recipe += '# [date] %s\n' %datetime.now().strftime('%b-%d-%Y')
    bot.info('+ ' + 'apprecipe '.ljust(5) + app)
    print(recipe_file)

    for section_name, section_content in config.items():
        content = "%s\n" %'\n'.join(section_content)
        header = '%' + section_name + ' %s' %app
        recipe += '%s\n%s\n' %(header, content)

    return recipe

            
def preview_environment(self, app, settings, config):
    '''preview the environment section

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars

    '''
    content = ''
    if "appenv" in config:
        content = '\n'.join(config['appenv'])
        bot.info('+ ' + 'appenv '.ljust(5) + app)
        print(settings['appenv'])
        print(content)
    return content



'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.client.utils import parse_input_preferences
from scif.logger import bot
import sys
import json
import os


def main(args,parser,subparser):

    from scif.main import ScifRecipe

    # Inspect choices, r, e, l, a
    choices = ['r', 'e', 'l', 'a', 'f', 'i', 'h', 'dump',
               'labels', 'all', 'files', "help", "install",
               'environment', 'definition', 'runscript'] 

    attributes = [x for x in args.attributes if x in choices]

    write_recipe = False
    if "dump" in attributes:
        write_recipe = True
        attributes.pop(attributes.index('dump'))

    apps = [x for x in args.attributes if x not in attributes]
    
    # First instantiate the client
    client = ScifRecipe(quiet=True, writable=False)

    # Filter down to apps that we want
    if len(apps) == 0:
        apps = client.apps()

    if len(attributes) == 0:
        attributes = ['a']

    result = {}
    for app in apps:

        inspection = client.inspect(app, attributes)
        if len(inspection) > 0:
            result[app] = inspection 

    if write_recipe is True:
        for app, atts in result.items():
            for key,val in atts.items():
                print('%' + key)
                print('\n'.join(val) + '\n')
    else:
        print(json.dumps(result,
                         indent=4,
                         separators=(',', ': ')))



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.utils import run_command as run_cmd
from scif.logger import bot
import shlex
import sys
import os
import re

def get_parts(pair, default=None):
    '''pair is expected to be a string with some key, value, and this function
       will replace any "=" with a space, and then parse the remainder
       for a first and second part (key and value pair). 

       We likely will need to use regular expressions for pairs that have 
       spaces or equals that need to be maintained.

       Parameters
       ==========
       pair should be a single string, likely with a key and value split by 
            a space or equals sign.

    '''
    pair = pair.replace('=',' ')
    parts = pair.split(' ')
    src = parts[0].strip()
    if len(parts) == 1:
        dest = default
    else:
        dest = parts[1].strip()
    return src,dest


def run_command(self, cmd, spinner=True):
    '''run_install will run a command (a list) and wrap in a spinner. A 
       result (dict) with message and return code is returned. If the
       return value is not 0, an error is issed and we exit
    '''
    if spinner is True:
        bot.spinner.start()
    
    result = run_cmd(cmd)

    if spinner is True:
        bot.spinner.stop()
    retval = result['return_code']
    if retval != 0:
        bot.error('Return code %s' %retval)
        sys.exit(retval)
    return result



def parse_entrypoint(entry_point=None):
    '''parse entrypoint will return a list, where the first argument is the
       executable, followed by arguments
 
       Parameters
       ==========
       entry_point: the entry point for the application, is 
    '''

    if entry_point is None:
        from scif.defaults import SCIF_ENTRYPOINT as entry_point

    # We expect an [executable, arg1, arg2]
    if not isinstance(entry_point, list):
        entry_point = [entry_point]

    # Special characters in the entrypoint should be replaced
    #        [e] in the command or entrypoint: environment vars --> $
    #        [out] in the command or entrypoint: environment vars --> >
    #        [in] in the command or entrypoint: environment vars --> <
    #        [pipe] in the command or entrypoint: environment vars --> |

    entry_point= ' '.join(entry_point)
    entry_point = re.sub('\[e\]','$', entry_point)
    entry_point = re.sub('\[out\]','>', entry_point)
    entry_point = re.sub('\[in\]','<', entry_point)
    entry_point = re.sub('\[pipe\]','|', entry_point)
    entry_point = re.sub('\[append\]','|', entry_point)

    # Split into executable, arguments
    return shlex.split(entry_point)



from setuptools import setup, find_packages
import codecs
import os

################################################################################
# HELPER FUNCTIONS #############################################################
################################################################################

def get_lookup():
    '''get version by way of scif.version, returns a 
    lookup dictionary with several global variables without
    needing to import singularity
    '''
    lookup = dict()
    version_file = os.path.join('scif', 'version.py')
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup


# Read in requirements
def get_requirements(lookup=None, key='INSTALL_REQUIRES'):
    '''get_requirements reads in requirements and versions from
    the lookup obtained with get_lookup'''

    if lookup == None:
        lookup = get_lookup()

    install_requires = []
    for module in lookup[key]:
        module_name = module[0]
        module_meta = module[1]
        if "exact_version" in module_meta:
            dependency = "%s==%s" %(module_name,module_meta['exact_version'])
        elif "min_version" in module_meta:
            if module_meta['min_version'] == None:
                dependency = module_name
            else:
                dependency = "%s>=%s" %(module_name,module_meta['min_version'])
        install_requires.append(dependency)
    return install_requires



# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__)) 
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup['__version__']
NAME = lookup['NAME']
AUTHOR = lookup['AUTHOR']
AUTHOR_EMAIL = lookup['AUTHOR_EMAIL']
PACKAGE_URL = lookup['PACKAGE_URL']
KEYWORDS = lookup['KEYWORDS']
DESCRIPTION = lookup['DESCRIPTION']
LICENSE = lookup['LICENSE']
with open('README.md') as filey:
    LONG_DESCRIPTION = filey.read()

################################################################################
# MAIN #########################################################################
################################################################################


if __name__ == "__main__":

    INSTALL_REQUIRES = get_requirements(lookup)
    INSTALL_REQUIRES_ALL = get_requirements(lookup,'INSTALL_REQUIRES_ALL')

    setup(name=NAME,
          version=VERSION,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          maintainer=AUTHOR,
          maintainer_email=AUTHOR_EMAIL,
          packages=find_packages(), 
          include_package_data=True,
          zip_safe=False,
          url=PACKAGE_URL,
          license=LICENSE,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          keywords=KEYWORDS,
          install_requires = INSTALL_REQUIRES,
          extras_require={
              'all': [INSTALL_REQUIRES_ALL]
          },
          classifiers=[
              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',
              'License :: OSI Approved :: MIT License',
              'Programming Language :: C',
              'Programming Language :: Python',
              'Topic :: Software Development',
              'Topic :: Scientific/Engineering',
              'Operating System :: Unix',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3',
          ],
          entry_points = {'console_scripts': [ 'scif=scif.client:main' ] })



'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
from scif.main.parser import ( load_filesystem, load_recipe )
import os
import sys

bot.level = 3

class ScifRecipe:
    '''Create and work with a scif recipe. Usage typically looks like:

       recipe = ScifRecipe('sregistry.scif')
       
       Parameters
       ==========
       path: the path to the scif recipe flie.

    '''
    
    def __init__(self, path=None, app=None, writable=True, quiet=False):
        '''initialize the scientific filesystem by creating a scif folder
           at the base, and loading the recipe to fill it.

           Parameters
           ==========
           path: is the first paramter, not required to initialize an empty 
                 client session. The logic proceeds as follows:
                 1. If path is not defined, we want (empty) interactive session
                 2. We derive the base from the environment SCIF_BASE
        '''

        # 0. base determined from environment
        from scif.defaults import SCIF_BASE
        self.set_defaults()

        # If recipe path not provided, try default base
        if path is None:
            path = SCIF_BASE

        # 1. Determine if path is a recipe or base
        if path is not None:

            self.set_base(SCIF_BASE, writable=writable) # /scif
            self.load(path, app, quiet)                 # recipe, environment

        # 2. Neither, development client
        else:
            bot.info('[skeleton] session!')
            bot.info('           load a recipe: client.load("recipe.scif")')
            bot.info('           change default base:  client.set_base("/")')


    def __str__(self):
        return '[scif]'

    def __repr__(self):
        return '[scif]'


    def speak(self):
        '''the client should announce self given that the shell is being used.
        '''
        if self._base is not None:
            apps = " | ".join(self.apps())
            bot.custom(prefix="%s %s" %(self, self._base), 
                                        message=apps, 
                                        color="CYAN")
        else:
            bot.info(self)


    def load(self, path, app=None, quiet=False):
        '''load a scif recipe into the object

        Parameters
        ==========
        path: the complete path to the config (recipe file) to load, or 
              root path of filesystem (that from calling function defaults to
              /scif)
        app:  if running with context of an active app, this will load the
              active app environment for it as well.
        '''
        # 1. path is a recipe
        if os.path.isfile(path):
            self._config = load_recipe(path)

        # 2. path is a base
        elif os.path.isdir(path):
            self._config = load_filesystem(path, quiet=quiet)

        else:
            bot.warning('%s is not detected as a recipe or base.' %path)
            self._config = None

        self.update_env(app)



'''

Copyright (C) 2017-2018 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017-2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
import sys
import pwd
import os


def main(args,parser,subparser):

    from scif.main import ScifRecipe
    apps = args.recipe

    if len(apps) > 0:
        recipe = apps.pop(0)

        if not os.path.exists(recipe):
            bot.error("Cannot find recipe file %s" %recipe)
            sys.exit(1)

        client = ScifRecipe(recipe, writable=False)

        # Preview the entire recipe, or the apps chosen
        client.preview(apps)

    else:
        bot.info('You must provide a recipe file to preview!')



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''


from scif.logger import bot
from scif.utils import ( mkdir_p, run_command, write_file, write_json )
from scif.main.helpers import get_parts
import sys
import os


def install(self, app=None):
    '''install recipes to a base. We assume this is the root of a system
       or container, and will write the /scif directory on top of it.
       If an app name is provided, install that app if it is found 
       in the config. This function goes through all step to:

       1. Install base folders to base, creating a folder for each app
       2. Install one or more apps to it, the config is already loaded
    '''

    self._install_base()             # Generate the folder structure
    self._install_apps(app)          # App install


def init_app(self, app):
    '''initialize an app, meaning adding the metadata folder, bin, and 
       lib to it. The app is created at the base
    '''
    settings = self.get_appenv_lookup(app)[app]

    # Create base directories for metadata
    for folder in ['appmeta', 'appbin', 'applib', 'appdata']:
        mkdir_p(settings[folder])
    return settings


def install_apps(self, apps=None):
    '''install one or more apps to the base. If app is defined, only
       install app specified. Otherwise, install all found in config.
    '''
    if apps in [None, '']:
        apps = self.apps()

    if not isinstance(apps, list):
        apps = [apps]

    if len(apps) == 0:
        bot.warning('No apps to install. Load a recipe or base with .load()')

    for app in apps:

        # We must have the app defined in the config
        if app not in self._config['apps']:
            bot.error('Cannot find app %s in config.' %app)
            sys.exit(1)

        # Make directories
        settings = self._init_app(app)

        # Get the app configuration
        config = self.app(app)

        # Get the app environment and export for install
        self.get_appenv(app, isolated=True, update=True)
        self.export_env(ps1=False)

        # Handle environment, runscript, labels
        self._install_runscript(app, settings, config)
        self._install_environment(app, settings, config)
        self._install_help(app, settings, config)
        self._install_labels(app, settings, config)
        self._install_files(app, settings, config)
        self._install_commands(app, settings, config)
        self._install_recipe(app, settings, config)
        self._install_test(app, settings, config)

        # After we install, in case interactive, deactivate last app
        self.deactivate(app)


def install_labels(self, app, settings, config):
    '''install labels will add labels to the app labelfile

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    lookup = dict()
    if "applabels" in config:
        labels = config['applabels']
        bot.level
        bot.info('+ ' + 'applabels '.ljust(5) + app)
        for line in labels:
            label, value = get_parts(line, default='')
            lookup[label] = value 
        write_json(lookup, settings['applabels'])
    return lookup


def install_files(self, app, settings, config):
    '''install files will add files (or directories) to a destination.
       If none specified, they are placed in the app base

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    if "appfiles" in config:
        files = config['appfiles']
        bot.info('+ ' + 'appfiles '.ljust(5) + app)

        for pair in files:
        
            # Step 1: determine source and destination
            src, dest = get_parts(pair, default=settings['approot'])

            # Step 2: copy source to destination
            cmd = ['cp']

            if os.path.isdir(src):
                cmd.append('-R')
            elif os.path.exists(src):
                cmd = cmd + [src, dest]
                result = self._run_command(cmd)
            else:    
                bot.warning('%s does not exist, skipping.' %src)



def install_commands(self, app, settings, config):
    '''install will finally, issue commands to install the app.

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    if "appinstall" in config:

        # Change directory so the APP is $PWD
        pwd = os.getcwd()
        os.chdir(settings['approot'])
        
        # issue install commands
        cmd = '\n'.join(config['appinstall'])
        bot.info('+ ' + 'appinstall '.ljust(5) + app)
        os.system(cmd)

        # Go back to previous location
        os.chdir(pwd)


def install_recipe(self, app, settings, config):
    '''Write the initial recipe for the app to its metadata folder.

       Parameters
       ==========
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    recipe_file = settings['apprecipe']
    recipe = '' 

    for section_name, section_content in config.items():
        content = '\n'.join(section_content)
        header = '%' + section_name
        recipe += '%s %s\n%s\n' %(header, app, content)

    write_file(recipe_file, recipe)
    return recipe


# Scripts

def install_script(self, section, app, settings, config):
    '''a general function used by install_runscript, install_help, and
       install_environment to write a script to a file from a config setting
       section

       Parameters
       ==========
       section: should be the name of the section in the config (e.g., apprun)
       app should be the name of the app, for lookup in config['apps']
       settings: the output of _init_app(), a dictionary of environment vars
       config: should be the config for the app obtained with self.app(app)

    '''
    if section in config:
        content = '\n'.join(config[section])
        bot.info('+ ' + section + ' '.ljust(5) + app)
        write_file(settings[section], content)


def install_runscript(self, app, settings, config):
    '''install runscript will prepare the runscript for an app.
       the parameters are shared by _install_script
    '''
    return self._install_script('apprun', app, settings, config)

            
def install_environment(self, app, settings, config):
    '''install will run the content to export environment variables, if defined
       the parameters are shared by _install_script
    '''
    return self._install_script('appenv', app, settings, config)


def install_help(self, app, settings, config):
    '''install will write the help section, if defined.
       the parameters are shared by _install_script
    '''
    return self._install_script('apphelp', app, settings, config)


def install_test(self, app, settings, config):
    '''install test will prepare a test script for an app.
       the parameters are shared by _install_script
    '''
    return self._install_script('apptest', app, settings, config)
